/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.domain;

import java.time.Duration;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.List;

import de.bsvrz.ibv.simbaustelle.migration.util.UndefiniertException;
import de.bsvrz.ibv.simbaustelle.parameter.IBssParameterObserver;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BaustellenSimulation;
import de.bsvrz.ibv.simbaustelle.verwaltung.BaustellenSchnittstelle;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Repräsentation für den Berechnungszeitraum einer Baustellen-Stauprognose.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class BssBerechnungsZeitRaum implements BerechnungsZeitRaum {

    /**
     * Logger für Debug-Ausgaben.
     */
    private static final Debug LOGGER = Debug.getLogger();

    /**
     * der Startzeitpunkt des Zeitraums.<br>
     * Der Startzeitpunkt ergibt sich aus dem Maximum des Tagesbeginns der
     * Gültigkeit der Baustelle und dem Tagesbeginn des Prognosezeitpunktes.
     */
    private LocalDateTime startZeitPunkt;

    /**
     * der Endzeitpunkt des Zeitraums.<br>
     * Der Endzeitpunkt ergibt sich aus dem Minimum von Tagesende des
     * Gültigkeitszeitraums der Baustelle und dem Tagesende des Prognosezeitraums
     * (-horizonts).
     */
    private LocalDateTime endZeitPunkt;

    /**
     * die Schrittweite in Millisekunden.<br>
     * Die Schrittweite wird als Parameter für die Baustellensimulation definiert.
     */
    private final Duration schrittWeite;

    /**
     * erzeugt eine Instanz des BerechnungsZeitraums auf Basis des übergebenen
     * Zeitpunkts, den Baustelleninformationen und den allgemeinen
     * Simulationsparametern.
     *
     * @param zeit      der Bezugszeitpunkt
     * @param baustelle die betroffene Baustelle
     * @param parameter die allgemeinen Simulationsparameter
     * @param manuell   Berechnung wurde manuell beauftragt
     * @throws UndefiniertException der Berechnungszeitraum konnte nicht bestimmt
     *                              werden, weil nicht alle Parameter verfügbar
     *                              waren
     */
    public BssBerechnungsZeitRaum(LocalDateTime zeit, final BaustellenSchnittstelle baustelle,
            final IBssParameterObserver parameter, final boolean manuell) throws UndefiniertException {

        if (manuell && (!BaustellenSimulation.getInstance().isZeitbereichAfoStrict())) {
            startZeitPunkt = LocalDateTime.of(baustelle.getDaten().getStart().toLocalDate(), LocalTime.MIDNIGHT);
            endZeitPunkt = LocalDateTime.of(baustelle.getDaten().getEnde().toLocalDate(), LocalTime.MIDNIGHT)
                    .plusDays(1);

            Duration manuellMsecs = BaustellenSimulation.getInstance().getMaxHorizontManuellMsecs();
            if (manuellMsecs != null && !manuellMsecs.isNegative() && !manuellMsecs.isZero()
                    && Duration.between(startZeitPunkt, endZeitPunkt).compareTo(manuellMsecs) > 0) {
                endZeitPunkt = startZeitPunkt.plus(manuellMsecs);
            }

        } else {

            LocalDateTime hCal = zeit;
            LocalDateTime bCal = baustelle.getDaten().getStart();

            // /Anfangszeitpunkt ermitteln
            LocalDateTime horizontZeit = LocalDateTime.of(hCal.toLocalDate(), LocalTime.MIDNIGHT);
            LOGGER.finer("Horizont-Startzeit: " + horizontZeit.format(DateTimeFormatter.ISO_LOCAL_DATE_TIME));

            LocalDateTime bstZeit = LocalDateTime.of(bCal.toLocalDate(), LocalTime.MIDNIGHT);
            LOGGER.finer(
                    "Baustellen-Startzeit: " + bstZeit.format(DateTimeFormatter.ISO_LOCAL_DATE_TIME));

            startZeitPunkt = horizontZeit;
            if (bstZeit.isAfter(startZeitPunkt)) {
                startZeitPunkt = bstZeit;
            }
            LOGGER.finer(
                    "Prognosestart: " + startZeitPunkt.format(DateTimeFormatter.ISO_LOCAL_DATE_TIME));

            // Endzeitpunkt ermitteln
            hCal = hCal.plus(parameter.getHorizont());
            horizontZeit = LocalDateTime.of(hCal.toLocalDate(), LocalTime.MIDNIGHT).plusDays(1);

            bCal = baustelle.getDaten().getEnde();
            bstZeit = LocalDateTime.of(bCal.toLocalDate(), LocalTime.MIDNIGHT).plusDays(1);

            endZeitPunkt = horizontZeit;
            if (bstZeit.isBefore(endZeitPunkt)) {
                endZeitPunkt = bstZeit;
            }
            LOGGER.finer(
                    "Prognoseende: " + endZeitPunkt.format(DateTimeFormatter.ISO_LOCAL_DATE_TIME));
        }

        schrittWeite = parameter.getPrognoseSchrittweite();
    }

    @Override
    public final LocalDateTime getEndeZeitpunkt() {
        return endZeitPunkt;
    }

    @Override
    public final List<LocalDateTime> getIntervalle() {
        final List<LocalDateTime> result = new ArrayList<>();
        LocalDateTime wert = startZeitPunkt;

        if (endZeitPunkt.isAfter(startZeitPunkt) && !schrittWeite.isNegative() && !schrittWeite.isZero()) {
            while (wert.isBefore(endZeitPunkt)) {
                result.add(wert);
                wert = wert.plus(schrittWeite);
            }
        }

        return result;
    }

    @Override
    public final LocalDateTime getStartZeitpunkt() {
        return startZeitPunkt;
    }

    @Override
    public boolean isValid() {
        return !startZeitPunkt.isAfter(endZeitPunkt);
    }

    @Override
    public String toString() {
        final StringBuffer buffer = new StringBuffer();
        buffer.append('(');
        buffer.append(startZeitPunkt);
        buffer.append('-');
        buffer.append(endZeitPunkt);
        buffer.append(", Schrittweite: ");
        buffer.append(schrittWeite);
        buffer.append(" ms)");
        return buffer.toString();
    }

    @Override
    public boolean prognoseMoeglich() {
        return endZeitPunkt.isAfter(startZeitPunkt);
    }

    @Override
    public Duration getSchrittweite() {
        return schrittWeite;
    }
}
