/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.prognose;

import java.time.Duration;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.concurrent.TimeUnit;

import de.bsvrz.ibv.simbaustelle.domain.BerechnungsZeitRaum;
import de.bsvrz.ibv.simbaustelle.domain.Prognose;
import de.bsvrz.ibv.simbaustelle.domain.PrognoseBaustelle;
import de.bsvrz.ibv.simbaustelle.domain.PrognoseSchritt;
import de.bsvrz.ibv.simbaustelle.domain.PrognostizierterStau;
import de.bsvrz.ibv.simbaustelle.migration.util.UndefiniertException;
import de.bsvrz.ibv.simbaustelle.parameter.IBssParameterObserver;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BaustellenSimulation;
import de.bsvrz.ibv.simbaustelle.verwaltung.BaustellenSchnittstelle;
import de.bsvrz.ibv.simbaustelle.verwaltung.SimulationsAuftrag;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * berechnet die für eine Baustelle zu erwartetenden Staus.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public abstract class BaustellenSimulator {

	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * die Baustelle, für die die Berechnung erfolgen soll.
	 */
	private final BaustellenSchnittstelle baustelle;

	/**
	 * die Parameter für die Berechnung.<br>
	 * Das Objekt wird aus der übergeordneten Baustellenverwaltung übernommen.
	 */
	private final IBssParameterObserver parameter;

	/** der Name des Auftraggebers der Simulation. */
	private final SimulationsAuftrag auftrag;

	private BaustellenSimulation applikation;

	/**
	 * erzeugt eine Instanz der Klasse zur Berechnung von Staus in einer Baustelle.
	 * 
	 * @param applikation die Applikation zur Ausführung der Baustellensimulation
	 * @param auftrag     der auszuführende Simulationsauftrag
	 * @param parameter   die Parameter für die Berechnung aus der
	 *                    Baustellenverwaltung
	 */
	public BaustellenSimulator(BaustellenSimulation applikation, final SimulationsAuftrag auftrag,
			final IBssParameterObserver parameter) {

		LOGGER.finer("Erzeuge BaustellenSimulator für Baustelle: " + auftrag.getBaustelle().getName());

		this.applikation = applikation;
		this.baustelle = auftrag.getBaustelle();
		this.auftrag = auftrag;
		this.parameter = parameter;
	}

	protected SimulationsAuftrag getAuftrag() {
		return auftrag;
	}

	protected BaustellenSchnittstelle getBaustelle() {
		return baustelle;
	}

	protected IBssParameterObserver getParameter() {
		return parameter;
	}

	/**
	 * ermittelt die Liste der Staus, die innerhalb des Prognosezeitraums zu
	 * erwarten sind.
	 *
	 * @param zeitRaum   der Zeitraum, für den die Staus prognostiziert werden
	 *                   sollen
	 * @param zufVerkehr der zufließende Verkehr
	 * @return die Daten der berechneten Prognose
	 * @throws UndefiniertException ein für die Prognose erforderlicher Wert konnte
	 *                              nicht ermittelt werden, die Prognose wurde nicht
	 *                              ausgeführt.
	 */
	protected final Prognose prognoseErstellen(final BerechnungsZeitRaum zeitRaum,
			final ZufliessenderVerkehrSchnittstelle zufVerkehr) throws UndefiniertException {

		Prognose simulation = new Prognose(new PrognoseBaustelle(auftrag.getBaustelle().getPid()));

		final long stBeginn = parameter.getLaengeStauBeginn();
		final long stAufhebung = parameter.getLaengeStauAufhebung();
		long aktuelleLaenge = 0;
		long fzImStau = 0;
		PrognostizierterStau aktuellerStau = null;

		final Duration schrittweite = zeitRaum.getSchrittweite();

		LOGGER.finer(baustelle.getName() + ": Beginne Stauberechnung mit " + " Parameter 'LängeStaubeginn': " + stBeginn
				+ ", 'LängeStauAufhebung': " + stAufhebung + ", 'Schrittweite': " + schrittweite);

		Map<LocalDateTime, Long> zuflussWerte = berechneZuflussWerte(zeitRaum, zufVerkehr);

		for (final LocalDateTime zs : zeitRaum.getIntervalle()) {
			try {
				Long zufluss = zuflussWerte.get(zs);
				if (zufluss == null) {
					throw new UndefiniertException("Lücken in Ganglinien sind nicht erlaubt!");
				}

				Long restKapazitaet = auftrag.getRestKapazitaet();
				if (restKapazitaet == null) {
					restKapazitaet = baustelle.getDaten().getRestKapazitaet(zs);
				}

				LOGGER.finer(baustelle.getName() + ": " + zs.format(DateTimeFormatter.ISO_LOCAL_DATE_TIME)
						+ ": Zufluss: " + zufluss + ", Restkapazität: " + restKapazitaet);

				
				final double fzDiff = (zufluss - restKapazitaet) * ((double) schrittweite.getSeconds() / TimeUnit.HOURS.toSeconds(1));

				final int fahrstreifenAnzahl = baustelle.anzahlFahrStreifen(aktuelleLaenge);
				final long laengePkw = parameter.getLaengePkw();

				final long lDiff = (long) ((fzDiff / fahrstreifenAnzahl) * laengePkw);
				aktuelleLaenge = Math.max(0, aktuelleLaenge + lDiff);
				fzImStau = Math.max(0, fzImStau + (long) fzDiff);
				final Duration verlustZeit = Duration
						.ofMillis((long) ((double) fzImStau / restKapazitaet * TimeUnit.HOURS.toMillis(1)));

				simulation.addSchritt(
						new PrognoseSchritt(zs.minus(zeitRaum.getSchrittweite()),
								aktuelleLaenge, verlustZeit, fzImStau));

				LOGGER.finer(baustelle.getName() + ": " + zs.format(DateTimeFormatter.ISO_LOCAL_DATE_TIME)
						+ ": FzDiff: " + fzDiff + ", FzImStau: " + fzImStau + ", Verlustzeit: " + verlustZeit);

				if (aktuellerStau == null) {
					if (aktuelleLaenge > stBeginn) {
						LOGGER.finer(baustelle.getName() + ": " + zs.format(DateTimeFormatter.ISO_LOCAL_DATE_TIME)
								+ ": neuen Stau erzeugen: Länge: " + aktuelleLaenge + ", Verlustzeit: " + verlustZeit);
						aktuellerStau = new PrognostizierterStau(zs, aktuelleLaenge, verlustZeit);
						simulation.addStau(aktuellerStau);
					} else {
						LOGGER.finer(baustelle.getName() + ": " + zs.format(DateTimeFormatter.ISO_LOCAL_DATE_TIME)
								+ ": es wird kein Stau erzeugt, da ermittelte Staulänge (" + aktuelleLaenge
								+ "m) < Parameter 'LängeStaubeginn' (" + stBeginn + "m)");
					}
				} else {
					if (aktuelleLaenge < stAufhebung) {
						LOGGER.finer(baustelle.getName() + ": " + zs.format(DateTimeFormatter.ISO_LOCAL_DATE_TIME)
								+ ": Stau beenden: Länge (" + aktuelleLaenge + "m) < Parameter 'LängeStauAufhebung' ("
								+ stAufhebung + "m)");
						aktuellerStau.beenden(zs);
						aktuellerStau = null;
					} else {
						LOGGER.finer(baustelle.getName() + ": " + zs.format(DateTimeFormatter.ISO_LOCAL_DATE_TIME)
								+ ": Stau aktualisieren: Länge: " + aktuelleLaenge + ", Verlustzeit: " + verlustZeit);
						aktuellerStau.aktualisieren(zs, aktuelleLaenge, verlustZeit);
					}
				}
			} catch (final UndefiniertException ex) {
				if (!BaustellenSimulation.getInstance().isAllowGaps()) {
					throw new UndefiniertException(
							"Daten für den Prognosezeitraum sind unvollständig: " + ex.getLocalizedMessage(), ex);
				}
				if (aktuellerStau != null) {
					aktuellerStau.beenden(zs);
					aktuellerStau = null;
					aktuelleLaenge = 0;
					fzImStau = 0;
					simulation.addSchritt(
							new PrognoseSchritt(zs.minus(zeitRaum.getSchrittweite()),
									aktuelleLaenge, Duration.ZERO, fzImStau));
				}
			}
		}

		return simulation;
	}

	public SortedMap<LocalDateTime, Long> berechneZuflussWerte(BerechnungsZeitRaum zeitRaum,
			ZufliessenderVerkehrSchnittstelle zufVerkehr) {
		SortedMap<LocalDateTime, Long> result = new TreeMap<>();
		int gaps = 0;
		for (LocalDateTime zs : zeitRaum.getIntervalle()) {
			try {
				result.put(zs, zufVerkehr.getFzg(zs));
			} catch (UndefiniertException e) {
				result.put(zs, null);
				gaps++;
			}
		}

		if (gaps <= 0) {
			return result;
		}

		Set<LocalDateTime> gapTimes = new LinkedHashSet<>();
		
		boolean fill = true;
		while (fill) {
			LocalDateTime lastValid = null;
			LocalDateTime firstError = null;
			LocalDateTime nextValid = null;
			int errors = 0;
			Iterator<LocalDateTime> iterator = result.keySet().iterator();

			while (iterator.hasNext()) {
				LocalDateTime zs = iterator.next();
				Long value = result.get(zs);
				if (value == null) {
					if (firstError == null) {
						firstError = zs;
					}
					errors++;
				} else {
					if (firstError == null) {
						lastValid = zs;
					} else {
						nextValid = zs;
						break;
					}
				}
			}

			if (firstError == null) {
				fill = false;
				continue;
			}

			if( errors > BaustellenSimulation.getInstance().getMaxGapSize()) {
				for (Entry<LocalDateTime, Long> entry : result.entrySet()) {
					if (!entry.getKey().isBefore(firstError) && (nextValid == null || entry.getKey().isBefore(nextValid))) {
						entry.setValue((long) 0);
						gapTimes.add(entry.getKey());
					}
				}
				
				continue;
			}
			
			if (lastValid == null) {
				if (nextValid != null) {
					while (firstError.isBefore(nextValid)) {
						result.put(firstError, result.get(nextValid));
					}
				} else {
					fill = false;
				}
				continue;
			}

			if (nextValid == null) {
				for (Entry<LocalDateTime, Long> entry : result.entrySet()) {
					if (!entry.getKey().isBefore(firstError)) {
						entry.setValue(result.get(lastValid));
					}
				}
				continue;
			}

			double diff = (double) (result.get(nextValid) - result.get(lastValid)) / (errors + 1);
			int count = 1;
			for (Entry<LocalDateTime, Long> entry : result.entrySet()) {
				if (entry.getKey().isBefore(firstError) || !entry.getKey().isBefore(nextValid)) {
					continue;
				}
				entry.setValue(Math.round((result.get(lastValid) + (count * diff))));
				count++;
			}

		}

		for(LocalDateTime time : gapTimes) {
			result.put(time, null);
		}
		
		return result;
	}

	protected BaustellenSimulation getApplikation() {
		return applikation;
	}
}
