/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.prognose.mdm;

import java.time.Duration;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.util.ArrayList;
import java.util.List;

import de.bsvrz.ibv.simbaustelle.domain.BerechnungsZeitRaum;
import de.bsvrz.ibv.simbaustelle.migration.util.UndefiniertException;
import de.bsvrz.ibv.simbaustelle.verwaltung.SimulationsAuftrag;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Bestimmung des Berechungszeitraums für eine Mdm-Baustelle.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class MDMBerechnungsZeitRaum implements BerechnungsZeitRaum {

    private static final Debug LOGGER = Debug.getLogger();

    private LocalDateTime startZeitPunkt;
    private LocalDateTime endeZeitPunkt;
    private List<LocalDateTime> intervalle = new ArrayList<>();

    private Duration schrittWeiteInMsec;

    public MDMBerechnungsZeitRaum(SimulationsAuftrag auftrag) throws UndefiniertException {

        LocalDateTime horizontZeit = LocalDateTime.of(auftrag.getStartZeit().toLocalDate(), LocalTime.MIDNIGHT);
        LOGGER.finer("Auftrags-Startzeit: " + horizontZeit);

        LocalDateTime bstZeit = LocalDateTime.of(auftrag.getBaustelle().getDaten().getStart().toLocalDate(),
                LocalTime.MIDNIGHT);
        LOGGER.finer("Baustellen-Startzeit: " + bstZeit);

        startZeitPunkt = horizontZeit;
        if (bstZeit.isAfter(startZeitPunkt)) {
            startZeitPunkt = bstZeit;
        }
        LOGGER.finer("Prognosestart: " + startZeitPunkt);

        // Endzeitpunkt ermitteln
        horizontZeit = LocalDateTime.of(auftrag.getEndZeit().toLocalDate(), LocalTime.MIDNIGHT);
        LOGGER.finer("Auftrags-Endzeit: " + horizontZeit);

        bstZeit = LocalDateTime.of(auftrag.getBaustelle().getDaten().getEnde().toLocalDate(), LocalTime.MIDNIGHT);
        LOGGER.finer("Baustellen-Endzeit: " + bstZeit);

        endeZeitPunkt = horizontZeit;
        if (bstZeit.isBefore(endeZeitPunkt)) {
            endeZeitPunkt = bstZeit;
        }
        endeZeitPunkt = endeZeitPunkt.plusDays(1);
        LOGGER.finer("Prognoseende: " + endeZeitPunkt);

        schrittWeiteInMsec = auftrag.getSchrittWeiteInMsec();
        if (schrittWeiteInMsec.isZero()) {
            schrittWeiteInMsec = Duration.ofHours(1);
        }

        if (endeZeitPunkt.isAfter(startZeitPunkt)) {
            LocalDateTime schrittZeitPunkt = startZeitPunkt.plus(schrittWeiteInMsec);
            while (!endeZeitPunkt.isBefore(schrittZeitPunkt)) {
                intervalle.add(schrittZeitPunkt);
                schrittZeitPunkt = schrittZeitPunkt.plus(schrittWeiteInMsec);
            }
        }
    }

    @Override
    public LocalDateTime getEndeZeitpunkt() {
        return endeZeitPunkt;
    }

    @Override
    public List<LocalDateTime> getIntervalle() {
        return intervalle;
    }

    @Override
    public LocalDateTime getStartZeitpunkt() {
        return startZeitPunkt;
    }

    @Override
    public boolean isValid() {
        return prognoseMoeglich();
    }

    @Override
    public boolean prognoseMoeglich() {
        return endeZeitPunkt.isAfter(startZeitPunkt) && !intervalle.isEmpty();
    }

    @Override
    public String toString() {
        final StringBuffer buffer = new StringBuffer();
        buffer.append('(');
        buffer.append(startZeitPunkt);
        buffer.append('-');
        buffer.append(endeZeitPunkt);
        buffer.append(", Schrittweite: ");
        buffer.append(schrittWeiteInMsec);
        buffer.append(" ms)");
        return buffer.toString();
    }

    @Override
    public Duration getSchrittweite() {
        if (schrittWeiteInMsec.isNegative()) {
            return Duration.ZERO;
        }
        return schrittWeiteInMsec;
    }
}
