/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.2 Baustellensimulation
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.simbaustelle.prognose.mdm;

import java.text.MessageFormat;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.ibv.simbaustelle.domain.BerechnungsZeitRaum;
import de.bsvrz.ibv.simbaustelle.domain.Prognose;
import de.bsvrz.ibv.simbaustelle.migration.util.UndefiniertException;
import de.bsvrz.ibv.simbaustelle.parameter.IBssParameterObserver;
import de.bsvrz.ibv.simbaustelle.prognose.BaustellenSimulator;
import de.bsvrz.ibv.simbaustelle.prognose.ZufliessenderVerkehr;
import de.bsvrz.ibv.simbaustelle.prognose.ZufliessenderVerkehrSchnittstelle;
import de.bsvrz.ibv.simbaustelle.simbaustelle.BaustellenSimulation;
import de.bsvrz.ibv.simbaustelle.verwaltung.SimulationsAuftrag;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Simulator zur Berechnung und Publikation der Prognosedaten einer
 * Mdm-Baustelle.
 * 
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class MdmBaustellenSimulator extends BaustellenSimulator {

    private static final Debug LOGGER = Debug.getLogger();

    public MdmBaustellenSimulator(BaustellenSimulation applikation, SimulationsAuftrag auftrag,
            IBssParameterObserver parameter) {
        super(applikation, auftrag, parameter);
    }

    public void ermittleDaten(Data datenSatz, Data datenSatzDetail) {

        MdmPrognoseDaten daten = new MdmPrognoseDaten(datenSatz);
        MdmPrognoseDatenDetail detailDaten = new MdmPrognoseDatenDetail(datenSatzDetail);

        boolean simulationErfolgreich = false;
        String fehlerMeldung = null;

 //       BaustellenSchnittstelle baustelle = getBaustelle();
        Prognose simulation = null;

        if (getBaustelle().pruefeSimulierbarkeit(getAuftrag().isManuell())) {

            try {
                final BerechnungsZeitRaum zeitRaum = new MDMBerechnungsZeitRaum(getAuftrag());

                if (zeitRaum.isValid()) {
                    if (zeitRaum.prognoseMoeglich()) {
                        final ZufliessenderVerkehrSchnittstelle zufVerkehr = new ZufliessenderVerkehr(getApplikation(),
                                getBaustelle(), getAuftrag().getBerechnungsMq(), zeitRaum);
                        simulation = prognoseErstellen(zeitRaum, zufVerkehr);
                        LOGGER.finer("Liste der Stauobjekte für " + getBaustelle().getName() + " ermittelt - Anzahl: "
                                + simulation.getStaus().size());
                    } else {
                        LOGGER.info("Prognosestart- und -endzeit identisch");
                        fehlerMeldung = "Prognosestart- und -endzeit identisch";
                    }
                    simulationErfolgreich = true;
                } else {
                    fehlerMeldung = MessageFormat.format(
                            "Der Berechungszeitraum der Baustelle {0} ist ungültig ( {1} )", getBaustelle(), zeitRaum);
                    getApplikation().getBetriebsMeldungsSender().warnung(getBaustelle().getObjekt().getPid(),
                            fehlerMeldung);
                    LOGGER.warning(fehlerMeldung);
                }
            } catch (final UndefiniertException e) {
                fehlerMeldung = MessageFormat.format(
                        "Es konnten nicht alle erforderlichen Daten für die Stauprognose der Baustelle {0} ermittelt werden: {1}",
                        getBaustelle(), e.getMessage());
                getApplikation().getBetriebsMeldungsSender().warnung(getBaustelle().getObjekt().getPid(),
                        fehlerMeldung);
                LOGGER.warning(fehlerMeldung);
            }
        }

        daten.setSimulationErfolgreich(simulationErfolgreich);
        daten.setAuftragGeber(getAuftrag().getAuftragGeber());
        daten.setBemerkung(getAuftrag().getBemerkung());
        daten.setFehlerMeldung(fehlerMeldung);
        if (simulation != null) {
            daten.setStauListe(simulation.getStaus());
        }

        detailDaten.setSimulationErfolgreich(simulationErfolgreich);
        detailDaten.setAuftragGeber(getAuftrag().getAuftragGeber());
        detailDaten.setBemerkung(getAuftrag().getBemerkung());
        detailDaten.setFehlerMeldung(fehlerMeldung);
        if (simulation != null) {
            detailDaten.setSchrittListe(simulation.getSchritte());
        }
    }

}
